#! /usr/bin/python
# Copyright 2016 Cumulus Networks LLC, all rights reserved

try:
    import time
    import poe_base
except ImportError, e:
    raise ImportError (str(e) + 'required module not found')

#############################################################################
# POE chip specific base class, inherits from the base_poe class,
# provides acces to the following POE chips:
#   MicroSemi PD69200
#

''' offsets within the PD69200 communication packet
'''
PD69200_PKT_OFFSET_KEY=0
PD69200_PKT_OFFSET_SEQ=1
PD69200_PKT_OFFSET_TXSUB0=2
PD69200_PKT_OFFSET_TXSUB1=3
PD69200_PKT_OFFSET_TXSUB2=4
PD69200_PKT_OFFSET_TXDATA0=5
PD69200_PKT_OFFSET_TXDATA1=6
PD69200_PKT_OFFSET_TXDATA2=7
PD69200_PKT_OFFSET_TXDATA3=8
PD69200_PKT_OFFSET_TXDATA4=9
PD69200_PKT_OFFSET_TXDATA5=10
PD69200_PKT_OFFSET_TXDATA6=11
PD69200_PKT_OFFSET_TXDATA7=12
PD69200_PKT_OFFSET_CKSUM0=13
PD69200_PKT_OFFSET_CKSUM1=14

PD69200_PKT_OFFSET_RXDATA0=2
PD69200_PKT_OFFSET_RXDATA1=3
PD69200_PKT_OFFSET_RXDATA2=4
PD69200_PKT_OFFSET_RXDATA3=5
PD69200_PKT_OFFSET_RXDATA4=6
PD69200_PKT_OFFSET_RXDATA5=7
PD69200_PKT_OFFSET_RXDATA6=8
PD69200_PKT_OFFSET_RXDATA7=9
PD69200_PKT_OFFSET_RXDATA8=10
PD69200_PKT_OFFSET_RXDATA9=11
PD69200_PKT_OFFSET_RXDATA10=12

PD69200_PKT_LEN=15
PD69200_PKT_CKSUM_LEN=2

''' packet field codes
'''
PD69200_PKT_KEY_CMD=0x00
PD69200_PKT_KEY_PROG=0x01
PD69200_PKT_KEY_REQUEST=0x02
PD69200_PKT_KEY_TELEMETRY=0x03
PD69200_PKT_KEY_REPORT=0x52

PD69200_PKT_SUB_CHANNEL=0x05
PD69200_PKT_SUB_E2=0x06
PD69200_PKT_SUB_GLOBAL=0x07
PD69200_PKT_SUB_RESTORE_DEFAULTS=0x2d

PD69200_PKT_SUB1_EN_DIS_4PAIR=0x02
PD69200_PKT_SUB1_PRIORITY=0x0a
PD69200_PKT_SUB1_SUPPLY=0x0b
PD69200_PKT_SUB1_PORTFULLINIT=0x0c
PD69200_PKT_SUB1_PORTSTATUS=0x0e
PD69200_PKT_SUB1_SAVE_CONFIG=0x0f
PD69200_PKT_SUB1_VERSIONZ=0x1e
PD69200_PKT_SUB1_MEASUREMENTS=0x25
PD69200_PKT_SUB1_PORTSTATUS1=0x31
PD69200_PKT_SUB1_PORTSTATUS2=0x32
PD69200_PKT_SUB1_PORTSTATUS3=0x33
PD69200_PKT_SUB1_SYSTEM_STATUS=0x3d
PD69200_PKT_SUB1_TEMP_MATRIX=0x43
PD69200_PKT_SUB1_ACTIVE_MATRIX=0x44
PD69200_PKT_SUB1_PORTSTATUS4=0x47
PD69200_PKT_SUB1_PORTSTATUS5=0x48
PD69200_PKT_SUB1_RESET=0x55
PD69200_PKT_SUB1_INDIVIDUAL_MASK=0x56

PD69200_PKT_SUB2_MAIN=0x17
PD69200_PKT_SUB2_SW_VERSION=0x21
PD69200_PKT_SUB2_POWER_BUDGET=0x57
PD69200_PKT_SUB2_POWER_MGMT_MODE=0x5f
PD69200_PKT_SUB2_POWER_INFO=0x60

PD69200_PKT_SUB2_4PAIR=0x34

''' packet data definitions
'''
PD69200_PKT_DATA_CMDMASKBOTH=0xf0
PD69200_PKT_DATA_CMDDISABLE=0x00
PD69200_PKT_DATA_CMDENABLE=0x01

PD69200_PKT_DATA_AFMASK_AF=0x00
PD69200_PKT_DATA_AFMASK_AT=0x01
PD69200_PKT_DATA_AFMASK_POH=0x02

PD69200_PKT_DATA_PRIORITY_LOW=3
PD69200_PKT_DATA_PRIORITY_HIGH=2
PD69200_PKT_DATA_PRIORITY_CRITICAL=1

PD69200_PKT_DATA_PORT_DISABLED=0
PD69200_PKT_DATA_PORT_ENABLED=1

''' status codes for 'getAllPortsStatus' request
'''
PD69200_PORT_STATUS_OK_LEGACY=0x00
PD69200_PORT_STATUS_OK_AF=0x01
PD69200_PORT_STATUS_OK_AT=0x02
PD69200_PORT_STATUS_HW_DISABLED=0x08
PD69200_PORT_STATUS_DISABLED=0x1a
PD69200_PORT_DETECTION_IN_PROGRESS=0x1b
PD69200_PORT_INVALID_DEVICE=0x1c
PD69200_PORT_OVERUNDERLOAD=0x1d
PD69200_PORT_UNDERLOAD=0x1e
PD69200_PORT_OVERLOAD=0x1f
PD69200_PORT_POWER_BUDGET_EXCEEDED=0x20
PD69200_PORT_INTERNAL_FAULT=0x21
PD69200_PORT_BAD_CAP_DETECTED=0x25
PD69200_PORT_OVERVOLTAGE=0x2d
PD69200_PORT_UNDERVOLTAGE=0x2e
PD69200_PORT_SHORT=0x34
PD69200_PORT_OVERTEMPERATURE=0x35
PD69200_PORT_TOOHOT=0x36
PD69200_PORT_STATUS_OK_HPD=0x40
PD69200_PORT_OVERPOWER=0x41
PD69200_PORT_INVALIDCLASS=0x43


pd69200_priority_map = { PD69200_PKT_DATA_PRIORITY_LOW:poe_base.PORT_CFG_PRIORITY_LOW,
                         PD69200_PKT_DATA_PRIORITY_HIGH:poe_base.PORT_CFG_PRIORITY_HIGH,
                         PD69200_PKT_DATA_PRIORITY_CRITICAL:poe_base.PORT_CFG_PRIORITY_CRITICAL}

pd69200_port_error_map = {
    PD69200_PORT_INVALID_DEVICE:'invalid device',
    PD69200_PORT_OVERUNDERLOAD:'overload/underload',
    PD69200_PORT_UNDERLOAD:'underload',
    PD69200_PORT_OVERLOAD:'overload',
    PD69200_PORT_POWER_BUDGET_EXCEEDED:'power budget exceeded',
    PD69200_PORT_INTERNAL_FAULT:'internal fault',
    PD69200_PORT_OVERVOLTAGE:'over voltage',
    PD69200_PORT_UNDERVOLTAGE:'under voltage',
    PD69200_PORT_SHORT:'short',
    PD69200_PORT_OVERTEMPERATURE:'over temperature',
    PD69200_PORT_TOOHOT:'too hot',
    PD69200_PORT_OVERPOWER:'over power',
    PD69200_PORT_INVALIDCLASS:'invalid class',
    PD69200_PORT_BAD_CAP_DETECTED:'detection failed'
}

pd69200_port_ok_results = {
    PD69200_PORT_STATUS_OK_LEGACY:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69200_PORT_STATUS_OK_AF:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69200_PORT_STATUS_OK_AT:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69200_PORT_STATUS_OK_HPD:poe_base.PortStatusResult.PORT_STATUS_DELIVERING_POWER,
    PD69200_PORT_STATUS_HW_DISABLED:poe_base.PortStatusResult.PORT_STATUS_DISABLED,
    PD69200_PORT_STATUS_DISABLED:poe_base.PortStatusResult.PORT_STATUS_DISABLED,
    PD69200_PORT_DETECTION_IN_PROGRESS:poe_base.PortStatusResult.PORT_STATUS_SEARCHING,
    PD69200_PORT_POWER_BUDGET_EXCEEDED:poe_base.PortStatusResult.PORT_STATUS_DENIED,
}

pd69200_port_pd_type_map = {
    PD69200_PORT_STATUS_OK_LEGACY:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_LEGACY,
    PD69200_PORT_STATUS_OK_AF:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_IEEE8023AF,
    PD69200_PORT_STATUS_OK_AT:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_IEEE8023AT,
    PD69200_PORT_STATUS_OK_HPD:poe_base.PortFullStatusResult.PORT_STATUS_PD_TYPE_HIGH_POWER,
}

# port events
PD69200_INT_PORT_EVENT_MASK = 0x3f

# system/device events
PD69200_INT_DEVICE_OVERTEMP1 =  1 << 8
PD69200_INT_DEVICE_OVERTEMP2 =  1 << 9
PD69200_INT_INTERNAL_DEVICE_FAULT = 1 << 10
PD69200_INT_MAIN_VOLTAGE_ERROR = 1 << 13
PD69200_INT_SYS_EVENT_MASK = PD69200_INT_DEVICE_OVERTEMP1 | PD69200_INT_DEVICE_OVERTEMP2 |\
    PD69200_INT_INTERNAL_DEVICE_FAULT | PD69200_INT_MAIN_VOLTAGE_ERROR


class PoePD69200(poe_base.PoeBase):

    def __init__(self):
        poe_base.PoeBase.__init__(self)
        self.seq = 0x10

    def poe_system_init(self):
        pass

    def get_poe_system_status(self, incremental=False):
        self.log_debug('PoePD69200::get_poe_system_status(%s)' % (str(incremental)))
        result = PD69200GetSysStatusResult()
        request = self._new_request_str([PD69200_PKT_KEY_REQUEST,
                                         0x00,
                                         PD69200_PKT_SUB_GLOBAL,
                                         PD69200_PKT_SUB1_SYSTEM_STATUS])
        response = self._send_request(request)
        result.handle_response(response, PD69200GetSysStatusResult.INTERRUPTS_RESPONSE)
        cmds_and_ports = [ (PD69200_PKT_SUB1_PORTSTATUS1, list(range(11))),
                           (PD69200_PKT_SUB1_PORTSTATUS2, list(range(11, 22))),
                           (PD69200_PKT_SUB1_PORTSTATUS3, list(range(22, 26))),
                           (PD69200_PKT_SUB1_PORTSTATUS4, list(range(26, 37))),
                           (PD69200_PKT_SUB1_PORTSTATUS5, list(range(37, 48)))]
        port_results = {}
        for port_cmd, ports in cmds_and_ports:
            request = self._new_request_str([PD69200_PKT_KEY_REQUEST,
                                            0x00,
                                            PD69200_PKT_SUB_GLOBAL,
                                            port_cmd])
            response = self._send_request(request)
            for port in ports:
                port_result = PD69200PortStatusResult(port)
                port_result.handle_response(response)
                port_results[port] = port_result._json()
        result.ports = port_results

        request = self._new_request_str([PD69200_PKT_KEY_REQUEST, 0x00,
                                       PD69200_PKT_SUB_GLOBAL,
                                       PD69200_PKT_SUB1_SUPPLY, PD69200_PKT_SUB2_MAIN])
        response = self._send_request(request)
        result.handle_response(response, PD69200GetSysStatusResult.SUPPLY_RESPONSE)
        return result._json()

    def get_port_extra_status(self, portList):
        self.log_debug('PoePD69200::get_port_extra_status(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        results = {}
        for port in portList:
            result = self._handle_port_command([PD69200_PKT_KEY_REQUEST, 0x00, PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PORTSTATUS, port],
                                             PD69200PortFullStatusResult(port),
                                             PD69200PortFullStatusResult.PORT_STATUS_RESPONSE)
            self._handle_port_command([PD69200_PKT_KEY_REQUEST, 0x00, PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_MEASUREMENTS, port],
                                             result, PD69200PortFullStatusResult.MEASUREMENTS_RESPONSE)
            self._handle_port_command([PD69200_PKT_KEY_REQUEST, 0x00, PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PRIORITY, port],
                                             result, PD69200PortFullStatusResult.PRIORITY_RESPONSE)
            results[port] = result._json()
        return results

    def get_port_priority(self, portList):
        self.log_debug('PoePD69200::get_port_priority(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        results = {}
        for port in portList:
            results[port] = self._handle_port_command([PD69200_PKT_KEY_REQUEST, 0x00, PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PRIORITY, port],
                                             PD69200PortPriorityResult(port))._json()
        return results

    def get_poe_diag_info(self):
        self.log_debug('PoePD69200::get_poe_diag_info')
        result = PD69200GetDiagInfoResult()

        ''' get power info
        '''
        request = self._new_request_str([PD69200_PKT_KEY_REQUEST, 0x00,
                                       PD69200_PKT_SUB_GLOBAL,
                                       PD69200_PKT_SUB1_SUPPLY, PD69200_PKT_SUB2_MAIN])
        response = self._send_request(request)
        result.handle_response(response, PD69200GetDiagInfoResult.SUPPLY_RESPONSE)

        ''' get power management info
        '''
        request = self._new_request_str([PD69200_PKT_KEY_REQUEST, 0x00,
                                         PD69200_PKT_SUB_GLOBAL,
                                         PD69200_PKT_SUB1_SUPPLY,
                                         PD69200_PKT_SUB2_POWER_MGMT_MODE])
        response = self._send_request(request)
        result.handle_response(response, PD69200GetDiagInfoResult.MODE_RESPONSE)

        ''' get individual mask info
        '''
        masks = (0x00, 0x01, 0x0c, 0x11, 0x14, 0x15, 0x1b, 0x1e, 0x1f, 0x28, 0x29, 0x2a, 0x2b, 0x2d, 0x2e, 0x2f,
                 0x30, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x43 )
        for mask in masks:
            request = self._new_request_str([PD69200_PKT_KEY_REQUEST, 0x00,
                                             PD69200_PKT_SUB_GLOBAL,
                                             PD69200_PKT_SUB1_INDIVIDUAL_MASK,
                                             mask])
            response = self._send_request(request)
            result.handle_response(response, PD69200GetDiagInfoResult.MASK_RESPONSE, mask)

        ''' get port info
        '''
        for port in range(48):
            port_result = self._handle_port_command([PD69200_PKT_KEY_REQUEST, 0x00, PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PORTSTATUS, port],
                                             PD69200GetPortDiagInfoResult(port),
                                             PD69200GetPortDiagInfoResult.STATUS_RESPONSE)

            self._handle_port_command([PD69200_PKT_KEY_REQUEST, 0x00, PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_SUPPLY, port],
                                             port_result, PD69200GetPortDiagInfoResult.POWER_LIMIT_RESPONSE)
            result.ports[port] = port_result._json()
        return result._json()

    def get_poe_version_info(self):
        self.log_debug('PoePD69200::get_poe_version_info')
        result = self._handle_sys_command([PD69200_PKT_KEY_REQUEST, 0x00,
                                           PD69200_PKT_SUB_GLOBAL,
                                           PD69200_PKT_SUB1_VERSIONZ,
                                           PD69200_PKT_SUB2_SW_VERSION],
                                           PD69200GetSwVersionResult())
        return result._json()

    def set_port_priority(self, portList, level):
        self.log_debug('PoePD69200::set_port_priority(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        priority = PD69200_PKT_DATA_PRIORITY_LOW
        for our_val, base_val in pd69200_priority_map.items():
            if base_val == level:
                priority = our_val
        for port in portList:
            request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                             PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PRIORITY, port,
                                             priority])
            self._send_request(request)

    def set_port_enable_poe(self, portList):
        self.log_debug('PoePD69200::set_port_enable_poe(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        for port in portList:
            request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                             PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PORTFULLINIT, port,
                                             PD69200_PKT_DATA_CMDMASKBOTH | PD69200_PKT_DATA_CMDENABLE,
                                             PD69200_PKT_DATA_AFMASK_AT])
            self._send_request(request)

    def set_port_disable_poe(self, portList):
        self.log_debug('PoePD69200::set_port_disable_poe(%s)' % (str(portList)))
        portList = self._mk_list(portList)
        for port in portList:
            request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                             PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PORTFULLINIT, port,
                                             PD69200_PKT_DATA_CMDMASKBOTH | PD69200_PKT_DATA_CMDDISABLE,
                                             PD69200_PKT_DATA_AFMASK_AT])
            self._send_request(request)

    def reset_ports(self, portList):
        self.log_debug('PoePD69200::reset_ports(%s)' % (str(portList)))
        ''' the PD69200 doesn't actually offer a 'reset port' function. We will
            do the best we can by disabling and then enabling the port.
        '''
        portList = self._mk_list(portList)
        for port in portList:
            result = self._handle_port_command([PD69200_PKT_KEY_REQUEST, 0x00, PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PORTSTATUS, port],
                                             PD69200PortFullStatusResult(port),
                                             PD69200PortFullStatusResult.PORT_STATUS_RESPONSE)
            if result.status == result.PORT_STATUS_DISABLED:
                continue

            request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                             PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PORTFULLINIT, port,
                                             PD69200_PKT_DATA_CMDMASKBOTH | PD69200_PKT_DATA_CMDDISABLE,
                                             PD69200_PKT_DATA_AFMASK_AT])
            self._send_request(request)
            time.sleep(0.1)
            request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                             PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_PORTFULLINIT, port,
                                             PD69200_PKT_DATA_CMDMASKBOTH | PD69200_PKT_DATA_CMDENABLE,
                                             PD69200_PKT_DATA_AFMASK_AT])
            self._send_request(request)

    ''' pd69200 specific extensions some of these are just for debug purposes
    '''
    def set_power_bank(self, bank, power_limit):
        self.log_debug('PoePD69200::set_power_bank(%u, %u)' % (bank, power_limit))
        request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                         PD69200_PKT_SUB_GLOBAL,
                                         PD69200_PKT_SUB1_SUPPLY,
                                         PD69200_PKT_SUB2_POWER_BUDGET, bank, (power_limit / 256),
                                         power_limit & 0xff,
                                         0x2, 0x3a, 0x01, 0xb8, 0x01])
        self._send_request(request)

    def get_power_bank_settings(self):
        for pb in range(16):
            request = self._new_request_str([PD69200_PKT_KEY_REQUEST, 0x00,
                                             PD69200_PKT_SUB_GLOBAL,
                                             PD69200_PKT_SUB1_SUPPLY,
                                             PD69200_PKT_SUB2_POWER_BUDGET, pb])
            self._send_request(request)

    def set_port_map(self):
        port_map = {
             0:(7, 6),     1:(5, 4),    2:(3, 2),    3:(1, 0),    4:(15, 14),
             5:(13, 12),   6:(11, 10),  7:(9, 8),    8:(23, 22),  9:(21, 20),
            10:(19, 18),  11:(17, 16), 12:(31, 30), 13:(29, 28), 14:(27, 26),
            15:(25, 24),  16:(39, 38), 17:(37, 36), 18:(35, 34), 19:(33, 32),
            20:(47, 46),  21:(45, 44), 22:(43, 42), 23:(41, 40), 24:(55, 54),
            25:(53, 52),  26:(51, 50), 27:(49, 48), 28:(63, 62), 29:(61, 60),
            30:(59, 58),  31:(57, 56), 32:(71, 70), 33:(69, 68), 34:(67, 66),
            35:(65, 64),  36:(79, 78), 37:(77, 76), 38:(75, 74), 39:(73, 72),
            40:(87, 86),  41:(85, 84), 42:(83, 82), 43:(81, 80), 44:(95, 94),
            45:(93, 92),  46:(91, 90), 47:(89, 88)
        }
        ''' enable system 4pair '''
        request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                         PD69200_PKT_SUB_GLOBAL,
                                         PD69200_PKT_SUB1_INDIVIDUAL_MASK,
                                         PD69200_PKT_SUB2_4PAIR,
                                         PD69200_PKT_DATA_PORT_ENABLED])
        self._send_request(request)
        ''' save settings '''
        self._persist_settings()

        ''' configure temp matrix '''
        for port in port_map:
            request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                             PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_TEMP_MATRIX,
                                             port,
                                             port_map.get(port)[0],
                                             port_map.get(port)[1]])
            self._send_request(request)

        ''' set active matrix '''
        request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                         PD69200_PKT_SUB_GLOBAL,
                                         PD69200_PKT_SUB1_TEMP_MATRIX])
        self._send_request(request)

        ''' enable/disable port 4pair '''
        en_dis = PD69200_PKT_DATA_PORT_ENABLED
        for port in range(48):
            if port == 12:
                en_dis = PD69200_PKT_DATA_PORT_DISABLED
            request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                             PD69200_PKT_SUB_CHANNEL,
                                             PD69200_PKT_SUB1_EN_DIS_4PAIR,
                                             port,
                                             en_dis])
            self._send_request(request)

        request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                         PD69200_PKT_SUB_GLOBAL,
                                         PD69200_PKT_SUB1_ACTIVE_MATRIX])
        self._send_request(request)
        self._persist_settings()

    def get_individual_masks(self):
        masks = (0x00, 0x01, 0x0c, 0x11, 0x14, 0x15, 0x1b, 0x1e, 0x1f, 0x28, 0x29, 0x2a, 0x2b, 0x2d, 0x2e, 0x2f,
                 0x30, 0x32, 0x33, 0x34, 0x35, 0x36, 0x37, 0x38, 0x39, 0x3a, 0x3b, 0x3c, 0x3d, 0x43 )
        for mask in masks:
            request = self._new_request_str([PD69200_PKT_KEY_REQUEST, 0x00,
                                             PD69200_PKT_SUB_GLOBAL,
                                             PD69200_PKT_SUB1_INDIVIDUAL_MASK,
                                             mask])
            result = self._send_request(request)

    def set_individual_mask(self):
        request = self._new_request_str([PD69200_PKT_KEY_CMD, 0x00,
                                         PD69200_PKT_SUB_GLOBAL,
                                         PD69200_PKT_SUB1_INDIVIDUAL_MASK,
                                         0x01, 0])
        result = self._send_request(request)

    ''' local utility routines
    '''
    def _persist_settings(self):
        request = self._new_request_str([PD69200_PKT_KEY_PROG, 0x00, PD69200_PKT_SUB_E2,
                                        PD69200_PKT_SUB1_SAVE_CONFIG])
        self._send_request(request)

    def _handle_sys_command(self, cmd_list, result):
        self.log_debug('PoePD69200::_handle_sys_command()')
        request = self._new_request_str(cmd_list)
        response = self._send_request(request)
        result.handle_response(response)
        return result

    def _handle_port_command(self, cmd_list, result, result_type=None):
        self.log_debug('PoePD69200::_handle_port_command()')
        request = self._new_request_str(cmd_list)
        response = self._send_request(request)
        result.handle_response(response, result_type)
        return result

    def _new_request_str(self, cmd_list, no_sequence=False):
        newRequest = []
        for i in range(len(cmd_list)):
            newRequest.append(cmd_list[i])
        for i in range(PD69200_PKT_LEN - PD69200_PKT_CKSUM_LEN - len(newRequest)):
            newRequest.append(0xff)
        if not no_sequence:
            newRequest[PD69200_PKT_OFFSET_SEQ] = self.seq
            self.seq += 1
            if self.seq == 0xff:
                self.seq = 0
        return newRequest

    def _send_request(self, request):
        retries = 3
        request += self._split_bytes(self._calc_cksum(request))
        while retries:
            try:
                self.lock()
                self._send(request)
                response = self._get_response()
                self.unlock()
                self._verify_response(response, request)
                return response
            except poe_base.POEDRuntimeError, e:
                retries -= 1
        raise e

    def _send(self, request):
        self.log_verbose('tx > %s' % (self._bytes_to_str(request)))
        rlen = len(request)
        if rlen != PD69200_PKT_LEN:
            raise poe_base.POEDRuntimeError('Attempt to send invalid length request - expecting %u, got %u' % (PD69200_PKT_LEN, rlen))
        self.poe_write(request)

    def _get_response(self):
        cstr = self.poe_read(PD69200_PKT_LEN)
        self.log_verbose('rx < %s' % (self._bytes_to_str(cstr)))
        return cstr

    def _clear_rx_queue(self):
        self.poe_read(PD69200_PKT_LEN)

    def _verify_response(self, response, request):
        if len(response) != PD69200_PKT_LEN:
            raise poe_base.POEDRuntimeError('invalid response length')
        elif response[PD69200_PKT_OFFSET_SEQ] != request[PD69200_PKT_OFFSET_SEQ]:
            self._clear_rx_queue()
            raise poe_base.POEDSequenceError('request/response sequence numbers don\'t match [0x%02X:0x%02X]\n' %
                                     (response[PD69200_PKT_OFFSET_SEQ], request[PD69200_PKT_OFFSET_SEQ]))
        calc_cksum = self._calc_cksum(response)
        rx_cksum = response[PD69200_PKT_OFFSET_CKSUM0] * 256 + response[PD69200_PKT_OFFSET_CKSUM1]
        if calc_cksum != rx_cksum:
            raise poe_base.POEDRuntimeError('invalid checksum in response [0x%02X:0x%02X]' % (calc_cksum, rx_cksum))

    def _calc_cksum(self, request):
        rlen = len(request)
        exp_len = PD69200_PKT_LEN - PD69200_PKT_CKSUM_LEN
        if rlen < exp_len:
            raise poe_base.POEDRuntimeError("Calculating checksum on invalid length request - expecting %u, got %u" % (exp_len, rlen))

        sum = 0
        for i in range(exp_len):
            sum += request[i]
        return sum

    def _split_bytes(self, bytes):
        return [bytes/256, bytes & 0xff]


class PD69200PortStatusResult(poe_base.PortStatusResult):

    def handle_response(self, response):
        if self.poe_port < 22:
            port_byte = response[PD69200_PKT_OFFSET_RXDATA0 + self.poe_port % 11]
        elif self.poe_port < 24:
            port_byte = response[PD69200_PKT_OFFSET_RXDATA0 + (self.poe_port - 22)]
        elif self.poe_port < 26:
            port_byte = response[PD69200_PKT_OFFSET_RXDATA0 + (self.poe_port - 19)]
        else:
            port_byte = response[PD69200_PKT_OFFSET_RXDATA0 + (self.poe_port - 26) % 11]
        self.status = pd69200_port_ok_results.get(port_byte, poe_base.PortStatusResult.PORT_STATUS_FAULT)
        if self.status == poe_base.PortStatusResult.PORT_STATUS_FAULT:
            self._error_str = pd69200_port_error_map.get(port_byte, 'other [0x%02X]' % (port_byte))
        elif self.status == poe_base.PortFullStatusResult.PORT_STATUS_DELIVERING_POWER:
            self.pd_type = pd69200_port_pd_type_map.get(port_byte, port_byte)

class PD69200PortFullStatusResult(poe_base.PortFullStatusResult):
    MEASUREMENTS_RESPONSE=1
    PORT_STATUS_RESPONSE=2
    PRIORITY_RESPONSE=3

    def handle_response(self, response, response_type):
        if response_type == self.PORT_STATUS_RESPONSE:
            en_dis = response[PD69200_PKT_OFFSET_RXDATA0]
            if en_dis == PD69200_PKT_DATA_PORT_DISABLED:
                self.status = poe_base.PortFullStatusResult.PORT_STATUS_DISABLED
            else:
                port_byte = response[PD69200_PKT_OFFSET_RXDATA1]
                self.status = pd69200_port_ok_results.get(port_byte, poe_base.PortFullStatusResult.PORT_STATUS_FAULT)
                if self.status == poe_base.PortFullStatusResult.PORT_STATUS_FAULT:
                    self._error_str = pd69200_port_error_map.get(port_byte, 'other [0x%02X]' % (port_byte))
                elif self.status == poe_base.PortFullStatusResult.PORT_STATUS_DELIVERING_POWER:
                    self.pd_class = response[PD69200_PKT_OFFSET_RXDATA4]
                    self.pd_type = pd69200_port_pd_type_map.get(port_byte, port_byte)
        elif response_type == self.MEASUREMENTS_RESPONSE:
            self.current = self._mk_word_data(response[PD69200_PKT_OFFSET_RXDATA2], response[PD69200_PKT_OFFSET_RXDATA3])           # mA
            self.power = self._mk_word_data(response[PD69200_PKT_OFFSET_RXDATA4], response[PD69200_PKT_OFFSET_RXDATA5])/1000.0      # Watts
            self.voltage = self._mk_word_data(response[PD69200_PKT_OFFSET_RXDATA7], response[PD69200_PKT_OFFSET_RXDATA8])/10.0      # Volts
        elif response_type == self.PRIORITY_RESPONSE:
            self.priority = pd69200_priority_map.get(response[PD69200_PKT_OFFSET_RXDATA0])

class PD69200PortPriorityResult(poe_base.PortPriorityResult):
        def handle_response(self, response, type):
            self.priority = pd69200_priority_map.get(response[PD69200_PKT_OFFSET_RXDATA0])

class PD69200GetSysPowerResult(poe_base.PoeResult):

    def handle_response(self, response):
        self.sys_allocated_power = self._mk_word_data(response[PD69200_PKT_OFFSET_RXDATA0], response[PD69200_PKT_OFFSET_RXDATA1])
        self.sys_available_power = self._mk_word_data(response[PD69200_PKT_OFFSET_RXDATA8], response[PD69200_PKT_OFFSET_RXDATA9])

class PD69200GetSysStatusResult(poe_base.PoeSysStatusResult):
    INTERRUPTS_RESPONSE=1
    SUPPLY_RESPONSE=2
    POWER_RESPONSE=3

    def handle_response(self, response, response_type):
        if response_type == self.INTERRUPTS_RESPONSE:
            int_val = response[PD69200_PKT_OFFSET_RXDATA9] * 256 + response[PD69200_PKT_OFFSET_RXDATA10]
            if int_val & PD69200_INT_SYS_EVENT_MASK:
                if int_val & PD69200_INT_DEVICE_OVERTEMP1 or int_val & PD69200_INT_DEVICE_OVERTEMP2:
                    self._error_str = 'System Overtemperature'
                elif int_val & PD69200_INT_INTERNAL_DEVICE_FAULT:
                    self._error_str = 'Internal System Fault'
                elif int_val & PD69200_INT_MAIN_VOLTAGE_ERROR:
                    self._error_str = 'System Main Voltage Error'
                else:
                    self._error_str = 'Unknown System Error [0x%04X]' % (int_val)
            if int_val & PD69200_INT_PORT_EVENT_MASK:
                self._port_event = True
        elif response_type == self.SUPPLY_RESPONSE:
            self.sys_consumed_power = self._mk_word_data(response[PD69200_PKT_OFFSET_RXDATA0], response[PD69200_PKT_OFFSET_RXDATA1])
            self.sys_power_limit = self._mk_word_data(response[PD69200_PKT_OFFSET_RXDATA8], response[PD69200_PKT_OFFSET_RXDATA9])

class PD69200GetSwVersionResult(poe_base.SwVersionResult):

    def handle_response(self, response):
        self.product_num = response[PD69200_PKT_OFFSET_RXDATA2]
        sw_ver = response[PD69200_PKT_OFFSET_RXDATA3] * 256 + response[PD69200_PKT_OFFSET_RXDATA4]
        sw_ver_str = '%04u' % (sw_ver)
        self.major_version = int(sw_ver_str[:2])
        self.minor_version = int(sw_ver_str[2])
        self.patch = int(sw_ver_str[3])
        self.ver_str = '%02X.%02u.%u.%u' % (self.product_num, self.major_version, self.minor_version, self.patch)

class PD69200GetDiagInfoResult(poe_base.DiagInfoResult):
    SUPPLY_RESPONSE=1
    MODE_RESPONSE=2
    MASK_RESPONSE=3

    def __init__(self):
        poe_base.DiagInfoResult.__init__(self)
        self.individual_masks = {}
        self.ports = {}

    def handle_response(self, response, response_type, response_info=None):
        if response_type == self.SUPPLY_RESPONSE:
            self.supply_power_consumption = response[PD69200_PKT_OFFSET_RXDATA0] * 256 + response[PD69200_PKT_OFFSET_RXDATA1]
            self.supply_max_shutdown_voltage = response[PD69200_PKT_OFFSET_RXDATA2] * 256 + response[PD69200_PKT_OFFSET_RXDATA3]
            self.supply_min_shutdown_voltage = response[PD69200_PKT_OFFSET_RXDATA4] * 256 + response[PD69200_PKT_OFFSET_RXDATA5]
            self.supply_power_bank = response[PD69200_PKT_OFFSET_RXDATA7]
            self.supply_power_limit = response[PD69200_PKT_OFFSET_RXDATA8] * 256 + response[PD69200_PKT_OFFSET_RXDATA9]
        elif response_type == self.MODE_RESPONSE:
            self.mode_pm1 = response[PD69200_PKT_OFFSET_RXDATA0]
            self.mode_pm2 = response[PD69200_PKT_OFFSET_RXDATA1]
            self.mode_pm3 = response[PD69200_PKT_OFFSET_RXDATA2]
        elif response_type == self.MASK_RESPONSE:
            self.individual_masks['0x%02X' % response_info] = '%u' % (response[PD69200_PKT_OFFSET_RXDATA0])

class PD69200GetPortDiagInfoResult(poe_base.PoePortResult):
    STATUS_RESPONSE=1
    POWER_LIMIT_RESPONSE=2
    PRIORITY_RESPONSE=3

    def handle_response(self, response, response_type):
        if response_type == self.STATUS_RESPONSE:
            self.status_enabled = '0x%02X' % response[PD69200_PKT_OFFSET_RXDATA0]
            self.status_status = '0x%02X' % response[PD69200_PKT_OFFSET_RXDATA1]
            self.status_force_power = '0x%02X' % response[PD69200_PKT_OFFSET_RXDATA2]
            self.status_latch = '0x%02X' % response[PD69200_PKT_OFFSET_RXDATA3]
            self.status_class = '0x%02X' % response[PD69200_PKT_OFFSET_RXDATA4]
            self.status_af_at = '0x%02X' % response[PD69200_PKT_OFFSET_RXDATA7]
            self.status_4pair = '0x%02X' % response[PD69200_PKT_OFFSET_RXDATA8]
        elif response_type == self.POWER_LIMIT_RESPONSE:
            self.power_ppl = '0x%04X' % (response[PD69200_PKT_OFFSET_RXDATA0] * 256 + response[PD69200_PKT_OFFSET_RXDATA1])
            self.power_tppl = '0x%04X' % (response[PD69200_PKT_OFFSET_RXDATA2] * 256 + response[PD69200_PKT_OFFSET_RXDATA3])
